import matplotlib.pyplot as plt

from .. import CAPlotter
from ..palettes import GRAYS
from ..palettes import SEABORN
from . import MplPlotter


class MplCAPlotter(MplPlotter, CAPlotter):

    """Matplotlib plotter for Correspondence Analysis"""

    def row_column_principal_coordinates(self, row_principal_coordinates,
                                         column_principal_coordinates, axes, explained_inertia,
                                         show_row_points, show_row_labels, show_column_points,
                                         show_column_labels):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')
        ax.axhline(y=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)
        ax.axvline(x=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)

        row_pc = row_principal_coordinates.iloc[:, axes].copy()
        row_pc.columns = ('X', 'Y')
        col_pc = column_principal_coordinates.iloc[:, axes].copy()
        col_pc.columns = ('X', 'Y')

        if show_row_points:
            ax.scatter(row_pc['X'], row_pc['Y'], c=SEABORN['blue'], s=50,
                       label='Row principal coordinates')

        if show_row_labels:
            ax.scatter(row_pc['X'], row_pc['Y'], alpha=0,
                       label=None if show_row_points else 'Row principal coordinates')
            for _, row in row_principal_coordinates.iterrows():
                ax.annotate(row.name, (row[axes[0]], row[axes[1]]), color=SEABORN['blue'])

        if show_column_points:
            ax.scatter(col_pc['X'], col_pc['Y'], c=SEABORN['red'], s=50,
                       label='Column principal coordinates')

        if show_column_labels:
            ax.scatter(col_pc['X'], col_pc['Y'], alpha=0,
                       label=None if show_column_points else 'Column principal coordinates')
            for _, row in column_principal_coordinates.iterrows():
                ax.annotate(row.name, (row[axes[0]], row[axes[1]]), color=SEABORN['red'])

        ax.axis('equal')

        ax.set_title('Row and column principal coordinates')
        ax.set_xlabel('Component {} ({}%)'.format(axes[0], 100 * round(explained_inertia[axes[0]], 2)))
        ax.set_ylabel('Component {} ({}%)'.format(axes[1], 100 * round(explained_inertia[axes[1]], 2)))
        if show_row_points or show_column_points:
            ax.legend(loc='best')

        return fig, ax



"""Multiple Correspondence Analysis"""
import numpy as np
import pandas as pd

from . import util
from .ca import CA
from .plot.mpl.mca import MplMCAPlotter


class MCA(CA):

    """
    Args:
        dataframe (pandas.DataFrame): A dataframe where each column is a variable.
        n_components (int): The number of principal components that have to be computed. The lower
            `n_components` is, the lesser time the CA will take to compute.
        use_benzecri_rates (bool): Whether to use Benzecri rates to inflate the eigenvalues.
        plotter (str): The plotting backend used to build the charts. Can be any of: 'mpl'.
    """

    def __init__(self, dataframe, n_components=2, use_benzecri_rates=False, plotter='mpl'):

        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError('dataframe muse be a pandas.DataFrame')

        self.categorical_columns = pd.DataFrame()
        self.supplementary_columns = pd.DataFrame()
        self.supplementary_rows = pd.DataFrame()

        self.initial_dataframe = dataframe.copy(deep=True)
        self.use_benzecri_rates = use_benzecri_rates

        supplementary_rows = None # Todo
        supplementary_columns = None  # Todo

        self._filter(
            dataframe=dataframe,
            supplementary_row_names=supplementary_rows if supplementary_rows else [],
            supplementary_column_names=supplementary_columns if supplementary_columns else []
        )

        super(MCA, self).__init__(
            dataframe=pd.get_dummies(dataframe),
            n_components=n_components,
            plotter=plotter
        )

    def _set_plotter(self, plotter_name):
        self.plotter = {
            'mpl': MplMCAPlotter()
        }[plotter_name]

    def _filter(self, dataframe, supplementary_row_names, supplementary_column_names):

        # Extract the categorical columns
        self.categorical_columns = dataframe.select_dtypes(exclude=[np.number])

        # Extract the supplementary rows
        self.supplementary_rows = dataframe.loc[supplementary_row_names].copy()
        self.supplementary_rows.drop(supplementary_column_names, axis=1, inplace=True)

        # Extract the supplementary columns
        self.supplementary_columns = dataframe[supplementary_column_names].copy()
        self.supplementary_columns.drop(supplementary_row_names, axis=0, inplace=True)

        # Remove the the supplementary columns and rows from the dataframe
        dataframe.drop(supplementary_row_names, axis=0, inplace=True)
        dataframe.drop(supplementary_column_names, axis=1, inplace=True)

    @property
    def q(self):
        """The number of columns in the initial dataframe

        As opposed to `p` which is the number of columns in the indicator matrix of the initial
        dataframe.
        """
        return self.initial_dataframe.shape[1]

    @property
    def eigenvalues(self):
        """The eigenvalues associated to each principal component.

        The eigenvalues are obtained by squaring the singular values obtained from a SVD. If
        `use_benzecri_rates` is `True` then Benzécri correction is applied to each eigenvalue.

        Returns:
            List[float]: The eigenvalues ordered increasingly.
        """
        eigenvalues = super(CA, self).eigenvalues
        if self.use_benzecri_rates:
            return util.calculate_benzecri_correction(eigenvalues)
        return eigenvalues

    @property
    def column_correlations(self):
        """The column correlation ratios with each principal component.

        The correlations ratios are the inter-group variances divided by the sum of the inter-group
        and intra-group variance of the numerical values associated to each categorical column.

        Returns:
            pandas.DataFrame: A dataframe of shape (`q`, `k`) containing the Pearson
            correlations between the columns and the principal components.
        """
        return pd.DataFrame({
            column.name: [
                util.correlation_ratio(column.tolist(), principal_component)
                for _, principal_component in self.row_principal_coordinates.iteritems()
            ]
            for _, column in self.initial_dataframe.iteritems()
        }).T

    @property
    def total_inertia(self):
        """The total inertia."""
        return (self.n_columns - self.q) / self.q

    def plot_rows(self, axes=(0, 1), show_points=True, show_labels=False, color_by=None,
                  ellipse_outline=False, ellipse_fill=False):
        """Plot the row principal coordinates.

        Args:
            axes (List(int)): A list of length two indicating which row principal coordinates to
                display.
            show_points (bool): Whether or not to show a point for each row principal coordinate.
            show_labels (bool): Whether or not to show the name of each row principal coordinate.
            color_by (str): Indicates according to which categorical variable the information should
                be colored by.
            ellipse_outline (bool): Whether or not to display an ellipse outline around each class
                if `color_by` has been set.
            ellipse_fill (bool): Whether or not to display a filled ellipse around each class if
                `color_by` has been set.
        """

        # Get color labels
        if color_by is None:
            color_labels = None
        elif color_by not in self.initial_dataframe.columns:
            raise ValueError("Categorical column '{}' can not be found".format(color_by))
        else:
            color_labels = self.initial_dataframe[color_by]

        return self.plotter.row_principal_coordinates(
            axes=axes,
            principal_coordinates=self.row_principal_coordinates,
            supplementary_principal_coordinates=pd.DataFrame(
                columns=self.row_principal_coordinates.columns
            ), # To do
            explained_inertia=self.explained_inertia,
            show_points=show_points,
            show_labels=show_labels,
            color_labels=color_labels,
            ellipse_outline=ellipse_outline,
            ellipse_fill=ellipse_fill
        )

    def plot_rows_columns(self, axes=(0, 1), show_row_points=True, show_row_labels=False,
                          show_column_points=True, show_column_labels=False):
        """Plot the row and column principal coordinates.

        Args:
            axes (List(int)): A list of length two indicating which row principal coordinates to
                display.
            show_row_points (bool): Whether or not to show a point for each row principal
                coordinate.
            show_row_labels (bool): Whether or not to show the name of each row principal
                coordinate.
            show_column_points (bool): Whether or not to show a point for each column principal
                coordinate.
            show_column_labels (bool): Whether or not to show the name of each column principal
                coordinate.
        """
        return self.plotter.row_column_principal_coordinates(
            axes=axes,
            row_principal_coordinates=self.row_principal_coordinates,
            column_principal_coordinates=self.column_principal_coordinates,
            explained_inertia=self.explained_inertia,
            show_row_points=show_row_points,
            show_row_labels=show_row_labels,
            show_column_points=show_column_points,
            show_column_labels=show_column_labels
        )

    def plot_relationship_square(self, axes=(0, 1), show_labels=True):
        """Plot the relationship square between the initial columns and the row principal
        components.

        Args:
            axes (List(int)): A list of length two indicating which row principal components to
                display.
            show_labels (bool): Whether or not to show the name of each column.
        """
        return self.plotter.relationship_square(
            axes=axes,
            column_correlations=self.column_correlations,
            explained_inertia=self.explained_inertia,
            show_labels=show_labels
        )



import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

from .. import Plotter
from ..palettes import SEABORN


class MplPlotter(Plotter):

    def __init__(self):

        mpl.rcParams['lines.linewidth'] = 1.8 # line width in points
        mpl.rcParams['lines.markeredgewidth'] = 0.3 # the line width around the marker symbol
        mpl.rcParams['lines.markersize'] = 7  # markersize, in points

        mpl.rcParams['grid.alpha'] = 0.5 # transparency, between 0.0 and 1.0
        mpl.rcParams['grid.linestyle'] = '-' # simple line
        mpl.rcParams['grid.linewidth'] = 0.4 # in points

    def inertia(self, explained_inertia):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        ax.plot(explained_inertia, color=SEABORN['blue'], label='Normalized inertia')
        ax.plot(explained_inertia, 'o', color=SEABORN['cyan'])

        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))

        ax.margins(0.05)
        ax.set_ylim(ymax=1)

        ax.set_title('Component contributions to inertia')
        ax.set_xlabel('Component number')
        ax.set_ylabel('Normalized inertia')
        ax.legend(loc='best')

        return fig, ax

    def cumulative_inertia(self, cumulative_explained_inertia, threshold):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        # Plot threshold line
        ax.axhline(y=threshold, color=SEABORN['red'], label='Threshold',
                   linestyle='--')

        # Plot first value above threshold line
        try:
            index_above_threshold = [
                i >= threshold
                for i in cumulative_explained_inertia
            ].index(True)
            ax.axvline(x=index_above_threshold, color=SEABORN['green'],
                       label='First component above threshold',
                       linestyle='--')
        except ValueError:
            pass

        # Plot inertia percentages curve
        ax.plot(cumulative_explained_inertia, color=SEABORN['blue'],
                label='Normalized cumulative inertia')
        ax.plot(cumulative_explained_inertia, 'o', color=SEABORN['blue'])

        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))

        ax.margins(0.05, 0.15)
        ax.set_ylim(ymin=0)

        ax.set_title('Cumulative component contributions to inertia')
        ax.set_xlabel('Component number')
        ax.set_ylabel('Normalized cumulative inertia')
        ax.legend(loc='best')

        return fig, ax



import matplotlib.cm as cm
import matplotlib.colors as clr
import matplotlib.pyplot as plt

from ..palettes import SEABORN


def create_discrete_cmap(n):
    """Create an n-bin discrete colormap."""
    if n <= len(SEABORN):
        colors = list(SEABORN.values())[:n]
    else:
        base = plt.cm.get_cmap('Paired')
        color_list = base([(i + 1) / (n + 1) for i in range(n)])
        cmap_name = base.name + str(n)
        return base.from_list(cmap_name, color_list, n)
    return clr.ListedColormap(colors)


def add_color_bar(ax, cmap, labels):
    """Add a colorbar to an axis.

    Args:
        ax (AxesSubplot)
        cmap (Colormap): A prepaped colormap of size n.
        labels (list of str): A list of strings of size n.
    """
    norm = clr.BoundaryNorm(list(range(cmap.N+1)), cmap.N)
    smap = cm.ScalarMappable(norm=norm, cmap=cmap)
    smap.set_array([])
    cbar = plt.colorbar(smap, ax=ax)
    cbar.set_ticks([i + 0.5 for i in range(cmap.N)])
    cbar.set_ticklabels(labels)



"""
Work in progress.
"""


import pandas as pd

from . import util
from .mca import MCA


class FAMD(MCA):

    """Factor Analysis of Mixed Data"""

    def _build_indicator_matrix(self):
        """Build the indicator matrix by placing a "1" where a row takes a value for a column and
        a "0" when it doesn't."""

        indicator_matrix = pd.get_dummies(self.X)

        # Add the numerical columns after rescaling
        for col in self.numerical_columns.columns:
            indicator_matrix[col] = util.rescale(
                series=self.numerical_columns[col],
                new_min=0,
                new_max=1
            )

        return indicator_matrix



import numpy as np

from prince.svd import SVD


def build_ellipse(X, Y):
    """Construct ellipse coordinates from two arrays of numbers.

    Args:
        X (1D array_like)
        Y (1D array_like)

    Returns:
        float: The mean of `X`.
        float: The mean of `Y`.
        float: The width of the ellipse.
        float: The height of the ellipse.
        float: The angle of orientation of the ellipse.

    """
    x_mean = np.mean(X)
    y_mean = np.mean(Y)

    cov_matrix = np.cov(X, Y)
    svd = SVD(cov_matrix, k=2)

    chi_95 = np.sqrt(4.61) # 90% quantile of the chi-square distribution
    width = np.sqrt(cov_matrix[0][0]) * chi_95 * 2
    height = np.sqrt(cov_matrix[1][1]) * chi_95 * 2

    eigenvector = svd.V.T[0]
    angle = np.arctan(eigenvector[1] / eigenvector[0])

    return x_mean, y_mean, width, height, angle



import matplotlib as mpl
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt

from .. import PCAPlotter
from .. import util as plot_util
from ..palettes import GRAYS
from ..palettes import SEABORN
from . import MplPlotter
from . import util as mpl_util


class MplPCAPlotter(MplPlotter, PCAPlotter):

    """Matplotlib plotter for Principal Component Analysis"""

    def row_principal_coordinates(self, axes, principal_coordinates,
                                  supplementary_principal_coordinates, explained_inertia,
                                  show_points, show_labels, color_labels, ellipse_outline,
                                  ellipse_fill):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')
        ax.axhline(y=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)
        ax.axvline(x=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)

        data = principal_coordinates.iloc[:, axes].copy() # Active rows
        supp = supplementary_principal_coordinates.iloc[:, axes].copy() # Supplementary rows
        data.columns = ('X', 'Y')
        supp.columns = ('X', 'Y')

        # Choose colors and add corresponding colorbar if necessary
        if (color_labels is not None) and (show_points or show_labels or ellipse_outline or ellipse_fill):
            data['label'] = color_labels
            supp['label'] = color_labels
            group_by = data.groupby('label')
            if not supplementary_principal_coordinates.empty:
                group_by_supp = supp.groupby('label')
            labels = list(group_by.groups.keys())
            n_colors = len(labels)
            cmap = mpl_util.create_discrete_cmap(n_colors)
            colors = cmap(range(n_colors))
            mpl_util.add_color_bar(ax, cmap, labels)

        if show_points:
            if color_labels is not None:
                for (label, group), color in zip(group_by, colors):
                    ax.scatter(group['X'], group['Y'], s=50, color=color, label=label)
                if not supplementary_principal_coordinates.empty:
                    for (label, group), color in zip(group_by_supp, colors):
                        ax.scatter(group['X'], group['Y'], s=90, color=color, label=label,
                                   marker='*')
            else:
                ax.scatter(data['X'], data['Y'], s=50, color=SEABORN['blue'])

        if show_labels:
            ax.scatter(data['X'], data['Y'], alpha=0, label=None)
            if color_labels is not None:
                for (label, group), color in zip(group_by, colors):
                    for _, row in group.iterrows():
                        ax.text(x=row['X'], y=row['Y'], s=row.name, color=color)
                if not supplementary_principal_coordinates.empty:
                    for (label, group), color in zip(group_by_supp, colors):
                        for _, row in group.iterrows():
                            ax.text(x=row['X'], y=row['Y'], s=row.name, color=color)
            else:
                for _, row in data.iterrows():
                    ax.text(x=row['X'], y=row['Y'], s=row.name)
                if not supplementary_principal_coordinates.empty:
                    for _, row in supp.iterrows():
                        ax.text(x=row['X'], y=row['Y'], s=row.name)

        if (ellipse_outline or ellipse_fill) and color_labels is not None:
            for (label, group), color in zip(group_by, colors):
                x_mean, y_mean, width, height, angle = plot_util.build_ellipse(group['X'], group['Y'])
                ax.add_patch(mpl.patches.Ellipse(
                    (x_mean, y_mean),
                    width,
                    height,
                    angle=angle,
                    linewidth=2 if ellipse_outline else 0,
                    color=color,
                    fill=ellipse_fill,
                    alpha=0.2 + (0.3 if not show_points else 0) if ellipse_fill else 1
                ))

        if not supplementary_principal_coordinates.empty:
            active_legend = mlines.Line2D([], [], marker='.', linestyle='', color=GRAYS['dark'],
                                          markersize=14, label='Active rows')
            supp_legend = mlines.Line2D([], [], marker='*', linestyle='', color=GRAYS['dark'],
                                             markersize=14, label='Supplementary rows')
            ax.legend(handles=[active_legend, supp_legend])

        ax.set_title('Row principal coordinates')
        ax.set_xlabel('Component {} ({}%)'.format(axes[0], 100 * round(explained_inertia[axes[0]], 2)))
        ax.set_ylabel('Component {} ({}%)'.format(axes[1], 100 * round(explained_inertia[axes[1]], 2)))

        return fig, ax


    def correlation_circle(self, axes, column_correlations, supplementary_column_correlations,
                           explained_inertia, show_labels):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')
        ax.axhline(y=0, linestyle='--', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)
        ax.axvline(x=0, linestyle='--', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)

        # Plot the arrows and add text
        for _, row in column_correlations.iterrows():
            ax.annotate(
                row.name if show_labels else '',
                xy=(0, 0),
                xytext=(row[axes[0]], row[axes[1]]),
                arrowprops=dict(arrowstyle='<-', edgecolor='black')
            )

        if not supplementary_column_correlations.empty:
            for _, row in supplementary_column_correlations.iterrows():
                ax.annotate(
                    row.name if show_labels else '',
                    xy=(0, 0),
                    xytext=(row[axes[0]], row[axes[1]]),
                    arrowprops=dict(arrowstyle='<-', edgecolor='red')
                )
            # Add legend to distinguish active and supplementary columns
            active_legend = mpatches.Patch(color='black', label='Active columns')
            supp_legend = mpatches.Patch(color='red', label='Supplementary columns')
            plt.legend(handles=[active_legend, supp_legend])

        circle = plt.Circle((0, 0), radius=1, color=GRAYS['dark'], fill=False, lw=1.4)
        ax.add_patch(circle)

        ax.margins(0.01)
        ax.axis('equal')

        ax.set_title('Correlation circle')
        ax.set_xlabel('Component {} ({}%)'.format(axes[0], 100 * round(explained_inertia[axes[0]], 2)))
        ax.set_ylabel('Component {} ({}%)'.format(axes[1], 100 * round(explained_inertia[axes[1]], 2)))

        return fig, ax



import fbpca


class SVD():

    """This object contains the results from a Singular Value Decomposition. The results are created
    when the object is instanciated."""

    U = None
    s = None
    V = None

    def __init__(self, X, **kwargs):
        """Perform the Singular Value Decomposition (SVD) of a matrix `X` of shape `(n, p)`.

        Args:
            X (matrix): The matrix on which to perform the SVD. `X` can be a `pandas.DataFrame`,
                however the SVD will be faster with a pure `numpy` matrix which can be extracted
                from a `pandas.DataFrame` thanks to the `values` property.
            kwargs: see http://fbpca.readthedocs.io/en/latest/#fbpca.pca.

        Returns:
            matrix: The left eigenvectors of shape `(n, k)`, usually denoted `U`.
            array: The singular values (square roots of the eigenvalues) of shape `(k,)`, usually
                denoted `s`.
            matrix: The right eigenvectors of shape `(k, p)`, usually denoted `V`.
        """
        self.U, self.s, self.V = fbpca.pca(X, **kwargs)



"""Principal Component Analysis"""
import numpy as np
import pandas as pd

from . import util
from .base import Base
from .plot.mpl.pca import MplPCAPlotter
from .svd import SVD


class PCA(Base):

    """
    Args:
        dataframe (pandas.DataFrame): The columns containing categorical data will be removed from
            the dataframe provided by the user; they will be stored in a separate dataframe at
            initialization.
        n_components (int): The number of principal components that have to be computed. The lower
            `n_components` is, the lesser time the PCA will take to compute.
        scaled (bool): Whether or not to rescale each variable by subtracting it's mean to it and
            then dividing it by it's standard deviation. This is advised when variables are not of
            the same order of magnitude.
        supplementary_rows (List[int/str]): A list of rows that won't be used to compute the PCA.
            These rows can however be displayed together with the active rows on a row
            principal coordinates chart.
        supplementary_columns (List[str]): A list of columns that won't be used to compute the PCA.
            The columns can however be displayed together with the active columns on a column
            correlation chart.
        plotter (str): The plotting backend used to build the charts. Can be any of: 'mpl'.
    """

    def __init__(self, dataframe, n_components=2, scaled=True, supplementary_rows=None,
                 supplementary_columns=None, plotter='mpl'):

        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError('dataframe muse be a pandas.DataFrame')

        self.categorical_columns = pd.DataFrame()
        self.supplementary_columns = pd.DataFrame()
        self.supplementary_rows = pd.DataFrame()

        self._filter(
            dataframe=dataframe,
            supplementary_row_names=supplementary_rows if supplementary_rows else [],
            supplementary_column_names=supplementary_columns if supplementary_columns else []
        )
        super(PCA, self).__init__(
            dataframe=dataframe,
            k=n_components,
            plotter=plotter
        )
        self._set_plotter(plotter_name=plotter)

        self.scaled = scaled
        if self.scaled:
            mean = self.X.mean()
            std = self.X.std()
            self.X = (self.X - mean) / std
            if not self.supplementary_rows.empty:
                self.supplementary_rows -= mean
                self.supplementary_rows /= std

        self._compute_svd()

    def _compute_svd(self):
        self.svd = SVD(X=self.X.values, k=self.n_components)

    def _set_plotter(self, plotter_name):
        self.plotter = {
            'mpl': MplPCAPlotter()
        }[plotter_name]

    def _filter(self, dataframe, supplementary_row_names, supplementary_column_names):

        # Extract the categorical columns
        self.categorical_columns = dataframe.select_dtypes(exclude=[np.number])

        # Extract the supplementary rows
        self.supplementary_rows = dataframe.loc[supplementary_row_names].copy()
        self.supplementary_rows.drop(self.categorical_columns.columns, axis=1, inplace=True)

        # Extract the supplementary columns
        self.supplementary_columns = dataframe[supplementary_column_names].copy()
        self.supplementary_columns.drop(supplementary_row_names, axis=0, inplace=True)

        # Remove the categorical column and the supplementary columns and rows from the dataframe
        dataframe.drop(supplementary_row_names, axis=0, inplace=True)
        dataframe.drop(supplementary_column_names, axis=1, inplace=True)
        dataframe.drop(self.categorical_columns.columns, axis=1, inplace=True)

    @property
    def n_supplementary_rows(self):
        """The number of supplementary rows."""
        return self.supplementary_rows.shape[0]

    @property
    def n_supplementary_columns(self):
        """The number of supplementary columns."""
        return self.supplementary_columns.shape[1]

    @property
    def row_principal_coordinates(self):
        """The row principal coordinates.

        The row principal coordinates are obtained by projecting `X` on it's right eigenvectors.
        This is done by calculating the dot product between `X` and `X`'s right eigenvectors.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row principal
            coordinates.
        """
        return pd.DataFrame(data=self.X.dot(self.svd.V.T), index=self.X.index)

    @property
    def supplementary_row_principal_coordinates(self):
        """The supplementary row principal coordinates.

        The row principal coordinates are obtained by projecting the supplementary rows on the right
        eigenvectors of `X`.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n_supplementary_rows`, `k`) containing the
            supplementary row principal coordinates.
        """
        return pd.DataFrame(
            data=self.supplementary_rows.dot(self.svd.V.T),
            index=self.supplementary_rows.index
        )

    @property
    def row_standard_coordinates(self):
        """The row standard coordinates.

        The row standard coordinates are obtained by scaling/dividing each row projection by it's
        associated eigenvalue.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row standard
            coordinates.
        """
        return self.row_principal_coordinates.div(self.eigenvalues, axis='columns')

    @property
    def supplementary_row_standard_components(self):
        """The supplementary row standard coordinates.

        The supplementary row standard coordinates are obtained by scaling/dividing each
        supplementary row projection by it's associated eigenvalue.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n_supplementary_rows`, `k`) containing the
            supplementary row standard coordinates.
        """
        return self.supplementary_row_principal_coordinates.div(self.eigenvalues, axis='columns')

    @property
    def row_component_contributions(self):
        """The row component contributions.

        Each row contribution towards each principal component is equivalent to the amount of
        inertia it contributes. This is calculated by dividing the squared row coordinates by the
        eigenvalue associated to each principal component.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row component
            contributions.
        """
        squared_coordinates = np.square(self.row_principal_coordinates)
        return squared_coordinates.div(self.eigenvalues, axis='columns')

    @property
    def row_cosine_similarities(self):
        """The squared row cosine similarities.

        The row cosine similarities are obtained by calculating the cosine of the angle shaped by
        the row principal coordinates and the row principal components. This is calculated by
        squaring each row projection coordinate and dividing each squared coordinate by the sum of
        the squared coordinates, which results in a ratio comprised between 0 and 1 representing the
        squared cosine.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the squared row cosine
            similarities.
        """
        squared_coordinates = np.square(self.row_principal_coordinates)
        total_squares = squared_coordinates.sum(axis='columns')
        return squared_coordinates.div(total_squares, axis='rows')

    @property
    def supplementary_row_cosine_similarities(self):
        """The supplementary squared row cosine similarities.

        The supplementary row cosine similarities are obtained by calculating the cosine of the
        angle shaped by the supplementary row principal coordinates and the supplementary row
        principal components.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n_supplementary_rows`, `k`) containing the
            squared supplementary row cosine similarities.
        """
        squared_coordinates = np.square(self.supplementary_row_principal_coordinates)
        total_squares = squared_coordinates.sum(axis='columns')
        return squared_coordinates.div(total_squares, axis='rows')

    @property
    def column_correlations(self):
        """The column correlations with each principal component.

        Returns:
            pandas.DataFrame: A dataframe of shape (`p`, `k`) containing the Pearson
            correlations between the columns and the principal components.
        """
        row_pc = self.row_principal_coordinates
        return pd.DataFrame(
            data=([col.corr(pc) for _, pc in row_pc.iteritems()] for _, col in self.X.iteritems()),
            columns=row_pc.columns,
            index=self.X.columns
        )

    @property
    def supplementary_column_correlations(self):
        """The supplementary column correlations with each principal component.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n_supplementary_columns`, `k`) containing
            the Pearson correlations between the supplementary columns and the principal components.
        """
        row_pc = self.row_principal_coordinates
        return pd.DataFrame(
            data=(
                [
                    col.corr(pc) if col.dtype in ('int64', 'float64')
                                 else util.correlation_ratio(col, pc)
                    for _, pc in row_pc.iteritems()
                ]
                for _, col in self.supplementary_columns.iteritems()
            ),
            columns=row_pc.columns,
            index=self.supplementary_columns.columns
        )

    @property
    def total_inertia(self):
        """The total inertia.

        Obtained by summing up the variance of each column.

        Returns:
            float: The total inertia.
        """
        return np.sum(np.square(self.X.values))

    def plot_rows(self, axes=(0, 1), show_points=True, show_labels=False, color_by=None,
                  ellipse_outline=False, ellipse_fill=False):
        """Plot the row principal coordinates.

        Args:
            axes (List(int)): A list of length two indicating which row principal coordinates to
                display.
            show_points (bool): Whether or not to show a point for each row principal coordinate.
            show_labels (bool): Whether or not to show the name of each row principal coordinate.
            color_by (str): Indicates according to which categorical variable the information should
                be colored by.
            ellipse_outline (bool): Whether or not to display an ellipse outline around each class
                if `color_by` has been set.
            ellipse_fill (bool): Whether or not to display a filled ellipse around each class if
                `color_by` has been set.
        """

        # Get color labels
        if color_by is None:
            color_labels = None
        elif color_by not in self.categorical_columns.columns:
            raise ValueError("'{}' is not a categorial column".format(color_by))
        else:
            color_labels = self.categorical_columns[color_by]

        return self.plotter.row_principal_coordinates(
            axes=axes,
            principal_coordinates=self.row_principal_coordinates,
            supplementary_principal_coordinates=self.supplementary_row_principal_coordinates,
            explained_inertia=self.explained_inertia,
            show_points=show_points,
            show_labels=show_labels,
            color_labels=color_labels,
            ellipse_outline=ellipse_outline,
            ellipse_fill=ellipse_fill
        )

    def plot_correlation_circle(self, axes=(0, 1), show_labels=True):
        """Plot the Pearson correlations between the components and the original columns.

        Args:
            axes (List(int)): A list of length two indicating which row principal coordinates to
                display.
            show_labels (bool): Whether or not to show the name of each column.
        """
        return self.plotter.correlation_circle(
            axes=axes,
            column_correlations=self.column_correlations,
            supplementary_column_correlations=self.supplementary_column_correlations,
            explained_inertia=self.explained_inertia,
            show_labels=show_labels
        )



from itertools import groupby
from operator import itemgetter

import numpy as np


def calculate_benzecri_correction(eigenvalues):
    """Because MCA and FAMD clouds often have a high dimensionality, the variance rates of the first
    principal components may be quite low, which makes them hard to interpret. Benzécri (1977) and
    Greenacre (1990) proposed to use modified rates to better appreciate the relative importance of
    the principal components."""
    p = len(eigenvalues)
    return [(p / (p - 1) * (eig - 1 / p)) ** 2 if eig > 1 / p else 0 for eig in eigenvalues]


def rescale(series, new_min=0, new_max=1):
    """Rescale the values of a `pandas.Series` from `new_min` to `new_max` using the min-max
    method."""
    old_min = series.min()
    old_max = series.max()
    series = series.apply(
        lambda x: (x - old_min) / (old_max - old_min) * \
                  (new_max - new_min) + new_min
    )
    return series


def cosine_similarity(A, B):
    """Calculate the cosine similarity between two arrays.

    Args:
        A (1D array_like)
        B (1D array_like)

    Returns:
        float: The cosine similarity between `A` and `B`.
    """
    return np.dot(A, B) / (np.sqrt(np.dot(A, A)) * np.sqrt(np.dot(B, B)))


def correlation_ratio(A, B):
    """Calculate the correlation ratio between a categorical array and a numerical array one.

    Args:
        A (1D array_like): contains categorial data.
        B (1D array_like): contains numerical data.

    Returns:
        float: The correlation ratio between `A` and `B`.
    """
    groups = groupby(sorted(zip(A, B), key=itemgetter(0)), key=itemgetter(0))
    means = []
    variances = []
    sizes = []
    for _, group in groups:
        values = [x for _, x in group]
        means.append(np.mean(values))
        variances.append(np.var(values))
        sizes.append(len(values))

    n = sum(sizes) # Total number of values, is also equal to the length of `A` and `B`
    mean = np.mean(means) # Global mean
    # The inter-group variance is the weighted mean of the group means
    var_inter = sum((s * (m - mean) ** 2 for s, m in zip(sizes, means))) / n
    # The intra-group variance is the weighted variance of the group variances
    var_intra = sum((s * v for s, v in zip(sizes, variances))) / n
    return var_inter / (var_inter + var_intra)



from prince.ca import CA
from prince.famd import FAMD
from prince.mca import MCA
from prince.pca import PCA
from prince.svd import SVD

__title__ = 'prince'
__version__ = '0.2.0'
__author__ = 'Max Halford'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 Max Halford'



import matplotlib.pyplot as plt

from .. import MCAPlotter
from ..palettes import GRAYS
from ..palettes import SEABORN
from . import MplPlotter
from . import util as mpl_util
from .pca import MplPCAPlotter


class MplMCAPlotter(MplPlotter, MCAPlotter):

    """Matplotlib plotter for Multiple Correspondence Analysis"""

    row_principal_coordinates = MplPCAPlotter.row_principal_coordinates

    def row_column_principal_coordinates(self, axes, row_principal_coordinates,
                                         column_principal_coordinates, explained_inertia,
                                         show_row_points, show_row_labels, show_column_points,
                                         show_column_labels):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')
        ax.axhline(y=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)
        ax.axvline(x=0, linestyle='-', linewidth=1.2, color=GRAYS['dark'], alpha=0.6)

        if show_row_points:
            ax.plot(
                row_principal_coordinates.iloc[:, axes[0]],
                row_principal_coordinates.iloc[:, axes[1]],
                marker='+',
                linestyle='',
                ms=7,
                label='Row principal coordinates',
                color=GRAYS['dark']
            )

        if show_row_labels:
            for _, row in row_principal_coordinates.iterrows():
                ax.annotate(row.name, (row[axes[0]], row[axes[1]]))

        # Extract the prefixes from each column name for coloring
        if show_column_points or show_column_labels:
            data = column_principal_coordinates.iloc[:, axes].copy()
            data.columns = ('X', 'Y')
            # Only keep the prefix of each label
            data['label'] = column_principal_coordinates.index.to_series().apply(lambda x: x.split('_')[0])
            group_by = data.groupby('label')
            labels = list(group_by.groups.keys())
            n_colors = len(labels)
            cmap = mpl_util.create_discrete_cmap(n_colors)
            colors = cmap(range(n_colors))
            mpl_util.add_color_bar(ax, cmap, labels)

        if show_column_points:
            for (label, group), color in zip(group_by, colors):
                ax.scatter(group['X'], group['Y'], color=color, s=50, label=label)

        if show_column_labels:
            for (label, group), color in zip(group_by, colors):
                for _, row in group.iterrows():
                    ax.text(row['X'], row['Y'], s=row.name, color=color)

        ax.set_title('Row and column principal coordinates')
        ax.set_xlabel('Component {} ({}%)'.format(axes[0], 100 * round(explained_inertia[axes[0]], 2)))
        ax.set_ylabel('Component {} ({}%)'.format(axes[1], 100 * round(explained_inertia[axes[1]], 2)))

        return fig, ax

    def relationship_square(self, axes, column_correlations, explained_inertia, show_labels):
        fig, ax = plt.subplots()

        ax.grid('on')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        data = column_correlations.iloc[:, axes].copy()
        data.columns = ('X', 'Y')

        for _, row in data.iterrows():
            ax.scatter(row['X'], row['Y'], s=50, color=SEABORN['blue'])
            if show_labels:
                ax.text(row['X'], row['Y'], s=row.name)

        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

        ax.set_title('Relationship square')
        ax.set_xlabel('Component {} ({}%)'.format(axes[0], 100 * round(explained_inertia[axes[0]], 2)))
        ax.set_ylabel('Component {} ({}%)'.format(axes[1], 100 * round(explained_inertia[axes[1]], 2)))

        return fig, ax



from collections import OrderedDict


GRAYS = OrderedDict([
    ('light', '#bababa'),
    ('dark', '#404040')
])

SEABORN = OrderedDict([
    ('blue', '#4c72b0'),
    ('green', '#55a868'),
    ('red', '#c44e52'),
    ('yellow', '#ccb974'),
    ('purple', '#8172b2'),
    ('cyan', '#64b5cd')
])



"""Correspondance Analysis"""
import numpy as np
import pandas as pd

from .base import Base
from .plot.mpl.ca import MplCAPlotter
from .svd import SVD


class CA(Base):

    """
    Args:
        dataframe (pandas.DataFrame): A contingency table.
        n_components (int): The number of principal components that have to be computed. The lower
            `n_components` is, the lesser time the CA will take to compute.
        plotter (str): The plotting backend used to build the charts. Can be any of: 'mpl'.
    """

    def __init__(self, dataframe, n_components=2, plotter='mpl'):
        super(CA, self).__init__(
            dataframe=dataframe,
            k=n_components,
            plotter=plotter
        )
        self._set_plotter(plotter_name=plotter)

        self._compute_svd()

    def _compute_svd(self):
        self.svd = SVD(X=self.standardized_residuals, k=self.n_components)

    def _set_plotter(self, plotter_name):
        self.plotter = {
            'mpl': MplCAPlotter()
        }[plotter_name]

    @property
    def N(self):
        """The total number of occurences in `X`.

        Returns:
            int: The sum of each cell in `X`.
        """
        return np.sum(self.X.values)

    @property
    def P(self):
        """The relative contingency table.

        Obtained by dividing each cell in `X` by `N`.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `p`) of total sum 1.
        """
        return self.X / self.N


    @property
    def row_sums(self):
        """The sum of each row in `P`.

        Returns:
            pandas.Series: A series of length `n` representing the row values distribution.
        """
        return self.P.sum(axis='columns')

    @property
    def column_sums(self):
        """The sum of each column in `P`.

        Returns:
            pandas.Series: A series of length `p` representing the column values distribution.
        """
        return self.P.sum(axis='rows')

    @property
    def expected_frequencies(self):
        """The expected frequencies in case of total independence.

        Calculated by performing an outer product between the row sums and the column sums.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `p`) representing the row and column
            values distribution.
        """
        return pd.DataFrame(
            data=self.row_sums.values.reshape(-1, 1).dot(self.column_sums.values.reshape(1, -1)),
            index=self.P.index,
            columns=self.P.columns
        )

    @property
    def row_masses(self):
        """Diagonal matrix of row masses.

        Returns:
            numpy.ndarray: A two dimensional array of shape (`n`, `n`) where each diagonal value
            represents the weight of the matching row; the non-diagonal cells are equal to 0.
        """
        return np.diag(1 / np.sqrt(self.row_sums))

    @property
    def column_masses(self):
        """Diagonal matrix of column masses.

        Returns:
            numpy.ndarray: A two dimensional array of shape (`p`, `p`) where each diagonal value
            represents the weight of the matching column; the non-diagonal cells are equal to 0.
        """
        return np.diag(1 / np.sqrt(self.column_sums))

    @property
    def standardized_residuals(self):
        """The matrix of standardized residuals.

        Obtained by normalizing the differences between the relative contingency table and the
        expected table by the row and column masses. This is the matrix that is used to calculate
        the SVD for Correspondance Analysis.

        Returns:
            numpy.ndarray: A two dimensional array of shape (`n`, `p`).
        """
        residuals = (self.P - self.expected_frequencies).values
        return self.row_masses.dot(residuals).dot(self.column_masses)

    @property
    def row_standard_coordinates(self):
        """The row standard coordinates.

        The row standard coordinates are obtained by projecting the row masses on the left
        eigenvectors.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row standard
            coordinates.
        """
        return pd.DataFrame(
            data=self.row_masses.dot(self.svd.U),
            index=self.P.index
        )

    @property
    def row_principal_coordinates(self):
        """The row principal coordinates.

        The row principal coordinates are obtained by multiplying the row standard coordinates by
        the singular values.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row principal
            coordinates.
        """
        return pd.DataFrame(
            data=self.row_standard_coordinates.values.dot(np.diag(self.svd.s)),
            index=self.P.index
        )

    @property
    def row_component_contributions(self):
        """The row component contributions.

        Each row contribution towards each principal component is equivalent to the amount of
        inertia it contributes. This is calculated by dividing the scaled squared row coordinates by
        the eigenvalue associated to each principal component.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the row component
            contributions.
        """
        squared_coordinates = np.square(self.row_principal_coordinates)
        scaled_squared_coordinates = squared_coordinates.multiply(self.row_sums, axis='rows')
        return scaled_squared_coordinates.div(self.eigenvalues, axis='columns')

    @property
    def row_cosine_similarities(self):
        """The squared row cosine similarities.

        The row cosine similarities are obtained by calculating the cosine of the angle shaped by
        the row principal coordinates and the row principal components. This is calculated by
        squaring each row projection coordinate and dividing each squared coordinate by the sum of
        the squared coordinates, which results in a ratio comprised between 0 and 1 representing the
        squared cosine.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `k`) containing the squared row cosine
            similarities.
        """
        squared_coordinates = np.square(self.row_principal_coordinates)
        total_squares = squared_coordinates.sum(axis='columns')
        return squared_coordinates.div(total_squares, axis='rows')

    @property
    def row_profiles(self):
        """The row profiles.

        The row profiles are obtained by dividing each cell in the relative contingency table by the
        sum of the row it belongs to.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `p`) containing the row profiles.
        """
        return pd.DataFrame(
            data=np.diag(1 / self.row_sums).dot(self.P),
            index=self.P.index,
            columns=self.P.columns
        )

    @property
    def column_standard_coordinates(self):
        """The column standard coordinates.

        The column standard coordinates are obtained by projecting the column masses on the right
        eigenvectors.

        Returns:
            pandas.DataFrame: A dataframe of shape (`p`, `k`) containing the column standard
            coordinates.
        """
        return pd.DataFrame(
            data=self.column_masses.dot(self.svd.V.T),
            index=self.P.columns
        )

    @property
    def column_principal_coordinates(self):
        """The column principal coordinates.

        The column principal coordinates are obtained by multiplying the column standard coordinates
        by the singular values.

        Returns:
            pandas.DataFrame: A dataframe of shape (`p`, `k`) containing the column principal
            coordinates.
        """
        return pd.DataFrame(
            data=self.column_standard_coordinates.values.dot(np.diag(self.svd.s)),
            index=self.P.columns
        )

    @property
    def column_component_contributions(self):
        """The column component contributions.

        Each column contribution towards each principal component is equivalent to the amount of
        inertia it contributes. This is calculated by dividing the scaled squared column coordinates
        by the eigenvalue associated to each principal component.

        Returns:
            pandas.DataFrame: A dataframe of shape (`p`, `k`) containing the column component
            contributions.
        """
        squared_coordinates = np.square(self.column_principal_coordinates)
        scaled_squared_coordinates = squared_coordinates.multiply(self.column_sums, axis='rows')
        return scaled_squared_coordinates.div(self.eigenvalues, axis='columns')

    @property
    def column_cosine_similarities(self):
        """The squared column cosine similarities.

        The column cosine similarities are obtained by calculating the cosine of the angle shaped by
        the column principal coordinates and the column principal components. This is calculated by
        squaring each column projection coordinate and dividing each squared coordinate by the sum
        of the squared coordinates, which results in a ratio comprised between 0 and 1 representing
        the squared cosine.

        Returns:
            pandas.DataFrame: A dataframe of shape (`p`, `k`) containing the squared row cosine
            similarities.
        """
        squared_column_pc = np.square(self.column_principal_coordinates)
        total_squares = squared_column_pc.sum(axis='rows')
        return squared_column_pc.div(total_squares, axis='columns')

    @property
    def column_profiles(self):
        """The column profiles.

        The column profiles are obtained by dividing each cell in the relative contingency table by
        the sum of the column it belongs to.

        Returns:
            pandas.DataFrame: A dataframe of shape (`n`, `p`) containing the column profiles.
        """
        return pd.DataFrame(
            data=(np.diag(1 / self.column_sums).dot(self.P.T).T),
            index=self.P.index,
            columns=self.P.columns
        )

    @property
    def total_inertia(self):
        """The total inertia.

        Obtained by summing up the squared relative differences between the relative contingency
        table and the expected frequencies.

        Returns:
            float: The total inertia.
        """
        expected = self.expected_frequencies
        return np.sum((np.square(self.P - expected) / expected).values)

    def plot_rows_columns(self, axes=(0, 1), show_row_points=True, show_row_labels=False,
                          show_column_points=True, show_column_labels=False):
        """Plot the row and column principal coordinates.

        Args:
            axes (List(int)): A list of length two indicating which row principal coordinates to
                display.
            show_row_points (bool): Whether or not to show a point for each row principal
                coordinate.
            show_row_labels (bool): Whether or not to show the name of each row principal
                coordinate.
            show_column_points (bool): Whether or not to show a point for each column principal
                coordinate.
            show_column_labels (bool): Whether or not to show the name of each column principal
                coordinate.
        """
        return self.plotter.row_column_principal_coordinates(
            axes=axes,
            row_principal_coordinates=self.row_principal_coordinates,
            column_principal_coordinates=self.column_principal_coordinates,
            explained_inertia=self.explained_inertia,
            show_row_points=show_row_points,
            show_row_labels=show_row_labels,
            show_column_points=show_column_points,
            show_column_labels=show_column_labels
        )



#!/usr/bin/python

from setuptools import setup, find_packages

from prince import  __author__, __license__, __title__, __version__


setup(
    author=__author__,
    author_email='maxhalford25@gmail.com',
    description='Factor analysis for in-memory datasets in Python',
    install_requires=[
        'fbpca>=1.0',
        'matplotlib>=1.5',
        'pandas>=0.19.0'
    ],
    license=__license__,
    long_description=open('README.md', 'r').read(),
    name=__title__,
    packages=find_packages(exclude=['tests']),
    url='https://github.com/MaxHalford/Prince',
    version=__version__,
)



import numpy as np
import pandas as pd


class Base():

    """Base contains the common operations performed during any factor analysis."""

    def __init__(self, dataframe, k, plotter):

        if plotter not in 'mpl':
            raise ValueError('Unrecognized plotting backend; choose from: mpl')

        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError('dataframe muse be a pandas.DataFrame')

        self.plotter = None
        self.svd = None

        self.__x = dataframe
        self.__n, self.__p = dataframe.shape
        # Determine the number of components computed during SVD
        self.__k = self.__p if k == -1 else min(k, self.__p)

    def _compute_svd(self):
        raise NotImplementedError

    def _set_plotter(self, plotter_name):
        raise NotImplementedError

    @property
    def X(self):
        """The dataset that is used to perform the SVD."""
        return self.__x

    @X.setter
    def X(self, matrix):
        self.__x = matrix

    @property
    def n_rows(self):
        """The number of rows in `X`."""
        return self.__n

    @property
    def n_columns(self):
        """The number of columns in `X`."""
        return self.__p

    @property
    def n_components(self):
        """The number of principal components that are calculated."""
        return self.__k

    @property
    def total_inertia(self):
        """The total inertia can be obtained differently for each kind of component analysis."""
        raise NotImplementedError

    @property
    def eigenvalues(self):
        """The eigenvalues associated to each principal component.

        The eigenvalues are obtained by squaring the singular values obtained from a SVD.

        Returns:
            List[float]: The eigenvalues ordered increasingly.
        """
        return np.square(self.svd.s).tolist()

    @property
    def explained_inertia(self):
        """The percentage of explained inertia per principal component.

        The explained inertia is obtained by dividing each eigenvalue by the total inertia.

        Returns:
            List[float]: The explained inertias ordered increasingly.
        """
        return [eig / self.total_inertia for eig in self.eigenvalues]

    @property
    def cumulative_explained_inertia(self):
        """The cumulative percentage of explained inertia per principal component.

        Returns:
            List[float]: The cumulative explained inertias ordered increasingly.
        """
        return np.cumsum(self.explained_inertia).tolist()

    def plot_inertia(self):
        """Plot a Scree diagram of the explained inertia per column."""
        return self.plotter.inertia(explained_inertia=self.explained_inertia)

    def plot_cumulative_inertia(self, threshold=0.8):
        """Plot a Scree diagram of the cumulative explained inertia per column.

        Args:
            threshold (float): The threshold at which a vertical line should be drawn. This is
                useful for finding the number of principal components required to reach a certain
                amount of cumulative inertia.
        """
        return self.plotter.cumulative_inertia(
            cumulative_explained_inertia=self.cumulative_explained_inertia,
            threshold=threshold
        )



class Plotter():

    def inertia(self, explained_inertia):
        """Generate a Scree diagram displaying the inertia of each component.

        Args:
            explained_inertia (list(float))
        """
        raise NotImplementedError

    def cumulative_inertia(self, cumulative_explained_inertia, threshold):
        """Generate a Scree diagram displaying the cumulative inertia of each component.

        Args:
            cumulative_explained_inertia (list(float))
            threshold (float): The thresold at which a vertical line should be plotted.
        """
        raise NotImplementedError


class PCAPlotter():

    def row_principal_coordinates(self, axes, principal_coordinates,
                                  supplementary_principal_coordinates, explained_inertia,
                                  show_points, show_labels, color_labels, ellipse_outline,
                                  ellipse_fill):
        """Generate a plot displaying the row principal coordinates.

        Args:
            axes (list(int)): A list containing two distinct elements, each one indicating the index
                of a principal coordinates to plot.
            principal_coordinates (pandas.DataFrame)
            supplementary_principal_coordinates (pandas.DataFrame)
            explained_inertia (list(float))
            show_points (bool): Whether or not to display points.
            show_labels (bool): Whether or not to display labels above each point.
            color_labels (pandas.Series): A series containing ordinal values according to which each
                point and label will be colored.
            ellipse_outline (bool): Whether or not to draw the outline of an ellipse around each
                group defined with `color_labels`.
            ellipse_outline (bool): Whether or not to draw the filling of an ellipse around each
                group defined with `color_labels`.
        """
        raise NotImplementedError

    def correlation_circle(self, axes, column_correlations, supplementary_column_correlations,
                           explained_inertia, show_labels):
        """Generate a plot displaying the columns/component Pearson correlations.

        Args:
            axes (list(int)): A list containing two distinct elements, each one indicating the index
                of a principal coordinates to plot.
            column_correlations (pandas.DataFrame): A dataframe containing the correlations
                between each initial column and each row principal component.
            supplementary_column_correlations (pandas.DataFrame)
            explained_inertia (list(float))
            show_labels (bool): Indicates whether or not to display labels above each point.
        """
        raise NotImplementedError


class CAPlotter():

    def row_column_principal_coordinates(self, axes, row_principal_coordinates,
                                         column_principal_coordinates, explained_inertia,
                                         show_row_points, show_row_labels, show_column_points,
                                         show_column_labels):
        """Generate a plot displaying the row and column principal coordinates simultaneously.

        Args:
            axes (list(int)): A list containing two distinct elements, each one indicating the index
                of a principal coordinate to plot.
            row_principal_coordinates (pandas.DataFrame)
            column_principal_coordinates (pandas.DataFrame)
            explained_inertia (list(float))
            show_row_points (bool): Indicates whether or not to display points for the row
                principal_coordinates.
            show_row_labels (bool): Indicates whether or not to display labels above each row
                projection point.
            show_row_points (bool): Indicates whether or not to display points for the column
                principal_coordinates.
            show_row_labels (bool): Indicates whether or not to display labels above each column
                projection point.
        """
        raise NotImplementedError


class MCAPlotter():

    def row_column_principal_coordinates(self, axes, row_principal_coordinates,
                                         column_principal_coordinates, explained_inertia,
                                         show_row_points, show_row_labels, show_column_points,
                                         show_column_labels):
        """Generate a plot displaying the row and column principal coordinates simultaneously.

        Args:
            axes (list(int)): A list containing two distinct elements, each one indicating the index
                of a principal coordinate to plot.
            row_principal_coordinates (pandas.DataFrame)
            column_principal_coordinates (pandas.DataFrame)
            explained_inertia (list(float))
            show_row_points (bool): Indicates whether or not to display points for the row
                principal coordinates.
            show_row_labels (bool): Indicates whether or not to display labels above each row
                projection point.
            show_row_points (bool): Indicates whether or not to display points for the column
                principal coordinates.
            show_row_labels (bool): Indicates whether or not to display labels above each column
                projection point.
        """
        raise NotImplementedError

    def relationship_square(self, axes, column_correlations, explained_inertia, show_labels):
        """Generate a plit displaying the correlations ratios between the initial columns and the
        row principal components.

        Args:
            axes (list(int)): A list containing two distinct elements, each one indicating the index
                of a principal component to plot.
            column_correlations (pandas.DataFrame): A dataframe containing the correlations
                between each initial column and each row principal component.
            explained_inertia (list(float))
            show_labels (bool): Indicates whether or not to display labels above each point.
        """
        raise NotImplementedError

